from typing import Optional
from fire import Fire


def compare_checkpoints(
    task_name: Optional[str] = None,
    objective: Optional[str] = None,
):
    pass


if __name__ == "__main__":
    pass
